	processor   6502
	
ZP_SCREENHI = $10 ; the screen to write to (backbuffer)
ZP_CHARSETHI = $11 ; the charset to write to (backbuffer)
ZP_SPRITECHARSETWRITE = $12 ; 2 byte the current written to sprite charset char * 8
ZP_SPRITECHARSETCHAR = $14 ; the current sprite charset char
ZP_PROCEDURETEMP = $15 ; 8 byte temporary memory
ZP_SPRITEREAD = $24 ; 2 byte read pointer
ZP_SPRITEWRITE = $26 ; 2 byte write pointer
ZP_SPRITES = $28 ; 2 byte sprite pointer head adress
ZP_SPRITE = $2a ; 2 byte current sprite data read
ZP_READ = $2c ; 2 byte multi purpose read
ZP_WRITE = $2e ; 2 byte multi purpose write
ZP_SPRITECHARSETWRITE2 = $30 ; 2 byte 
ZP_SPRITECHARSETWRITE3 = $32 ; 2 byte 
ZP_SPRITECHARSETWRITE4 = $34 ; 2 byte 

SCREENROWS = 25
SCREENCOLS = 40

CHARSET1 = $1800
CHARSET2 = $2000
SCREEN1 = $2800
SCREEN2 = $3000

	ORG $1001
	dc.w $100B,0
	dc.b $9E,"4109",0,0,0	; SYS4109
	jmp MAINPRG

	ORG $4000
;--------------------------------------------------------------------
; Hauptprogramm (Init)
;--------------------------------------------------------------------
MAINPRG SUBROUTINE

	jsr genTables
	jsr initSprites
	jsr hardwareSetup
	jsr initScreen

.mainloop
	jsr switchFrame

	lda ZP_SCREENHI ; clear the background
	ora #$04
	clc
	sta .again2 + 2 + 3*0
	adc #$01
	sta .again2 + 2 + 3*1
	adc #$01
	sta .again2 + 2 + 3*2
	adc #$01
	sta .again2 + 2 + 3*3

	lda #$00
	ldx #$00
.again2
	sta $4400,x
	sta $4400,x
	sta $4400,x
	sta $4400,x
	inx
	bne .again2

	ldx #$00
.again
	lda #$01
	sta spriteVisible,x
	txa
	asl
	asl
	asl
	asl
	adc frameCounter
	tay
	lda sintab,y
	lsr
	adc #[200-128]/2
	sta spriteYPosLo,x
	txa
	asl
	asl
	asl
	asl
	asl
	clc
	adc #50
	adc frameCounter
	tay
	lda sintab,y
	sta spriteXPosLo,x
	txa
	asl
	adc frameCounter
	lsr
	lsr
	and #31
	cmp #24
	bcc .nosub
	sec
	sbc #24
.nosub
	sta spritePointer,x
	txa
	asl
	asl
	adc frameCounter
	lsr
	lsr
	lsr
	lsr
	lsr
	and #$01
	sta spriteXFlipped,x

	inx
	cpx #SPRITECOUNT/2
	bne .again

	jsr clearSprites ; does nothing currently
	jsr paintSprites

	inc frameCounter

	jmp .mainloop

doubleBuffer dc.b $00
frameCounter dc.b $00

;--------------------------------------------------------------------
;--------------------------------------------------------------------
switchFrame SUBROUTINE
	jsr waitUntilBorder
	lda doubleBuffer
	eor #$01
	sta doubleBuffer
	bne .zwo
	lda #>SCREEN1
	sta $ff14
	lda #>CHARSET1
	sta $ff13
	lda #>SCREEN2
	sta ZP_SCREENHI
	lda #>CHARSET2
	sta ZP_CHARSETHI
	jmp .lower
.zwo
	lda #>SCREEN2
	sta $ff14
	lda #>CHARSET2
	sta $ff13
	lda #>SCREEN1
	sta ZP_SCREENHI
	lda #>CHARSET1
	sta ZP_CHARSETHI
.lower
	; initialize char write to for sprites
	lda #FIRSTSPRITECHAR
	sta ZP_SPRITECHARSETCHAR
	lda #<[8*FIRSTSPRITECHAR]
	sta ZP_SPRITECHARSETWRITE + 0
	lda #>[8*FIRSTSPRITECHAR]
	clc
	adc ZP_CHARSETHI
	sta ZP_SPRITECHARSETWRITE + 1
	rts	
;--------------------------------------------------------------------
; initializes the screen with default data
;--------------------------------------------------------------------
initScreen SUBROUTINE
	ldx #$00
.again
	lda #$63 + $08
	sta SCREEN1 + $100 * 0,x
	sta SCREEN1 + $100 * 1,x
	sta SCREEN1 + $100 * 2,x
	sta SCREEN1 + $100 * 3,x
	sta SCREEN2 + $100 * 0,x
	sta SCREEN2 + $100 * 1,x
	sta SCREEN2 + $100 * 2,x
	sta SCREEN2 + $100 * 3,x
	lda #$00
	sta SCREEN1 + $100 * 0 + $400,x
	sta SCREEN1 + $100 * 1 + $400,x
	sta SCREEN1 + $100 * 2 + $400,x
	sta SCREEN1 + $100 * 3 + $400,x
	sta SCREEN2 + $100 * 0 + $400,x
	sta SCREEN2 + $100 * 1 + $400,x
	sta SCREEN2 + $100 * 2 + $400,x
	sta SCREEN2 + $100 * 3 + $400,x
	inx
	bne .again
	rts

;--------------------------------------------------------------------
; initializes screen and hardware
;--------------------------------------------------------------------
hardwareSetup

	; disable hardware cursor
	lda #>$03fe
	sta $ff0c
	lda #<$03fe
	sta $ff0d

	lda #$00
	sta $ff15
	sta $ff19

	;-------------------------------------------------------------------
	; bitmap adress 2000 | bitmap ram 
	;-------------------------------------------------------------------
	lda $ff12
	and #%11000011 
	ora #%00001000 
	sta $ff12

	lda #>SCREEN1
	sta $ff14

	lda $ff13
	and #%00000101 ; speedmode
	ora #>CHARSET1
	sta $ff13

	lda $ff07
	and #%01000000
	ora #%10010000 ; we want multicolor
	sta $ff07
	rts

waitFrame SUBROUTINE
	jsr waitUntilBorder
	jmp waitUntilFrame

waitUntilBorder SUBROUTINE
	lda #$01
.waitb
	bit $ff1c
	beq .waitb
	rts

waitUntilFrame SUBROUTINE
	lda #$01
.waitb
	bit $ff1c
	bne .waitb
	rts

	include "tables.asm"
	include "sprites.asm"
	include "sprites1.def"
spriteStart
	include "sprites1.inc"
sintab
	incbin "sintab.bin"

	echo "spriteCodeSize:", spriteCodeSizeEnd - spriteCodeSizeStart
	echo "tableSize:", tableSizeEnd - tableSizeStart
	echo "end of file: ", [.]
