SPRITECOUNT = 16

FIRSTSPRITECHAR = 128
FILLUPBYTE = $ff

rectLeft	dc.b $00
rectRight	dc.b $00
rectBottom	dc.b $00
rectTop		dc.b $00

scrollXLo dc.b $00
scrollXHi dc.b $00
scrollYLo dc.b $00
scrollYHi dc.b $00

; sprites to paint
spriteXPosLo		ds.b SPRITECOUNT,$00 ; map x position low byte
spriteXPosHi		ds.b SPRITECOUNT,$00 ; map x position hi byte
spriteYPosLo		ds.b SPRITECOUNT,$00 ; map y position low byte
spriteYPosHi		ds.b SPRITECOUNT,$00 ; map y position hi byte
spriteXFlipped		ds.b SPRITECOUNT,$00 ; should it be flipped in x
spriteVisible		ds.b SPRITECOUNT,$00 ; is it visible
spritePointer		ds.b SPRITECOUNT,$00 ; the graphics number of the sprite
spriteScreenSpace	ds.b SPRITECOUNT,$00 ; is it screenspace (so no scrolling)

; doublebuffered clear (0-SPRITECOUNT restores for buffer 0, SPRITECOUNT-SPRITECOUNT*2 restores for buffer 1)
spriteRestoreOn ds.b SPRITECOUNT*2,$00 ; should this sprite be restored
spriteRestoreXS ds.b SPRITECOUNT*2,$00 ; rectangle of the sprite in screen chars
spriteRestoreYS ds.b SPRITECOUNT*2,$00 ; rectangle of the sprite in screen chars
spriteRestoreXE ds.b SPRITECOUNT*2,$00 ; rectangle of the sprite in screen chars
spriteRestoreYE ds.b SPRITECOUNT*2,$00 ; rectangle of the sprite in screen chars
spriteRestoreIndexAdd
			dc.b SPRITECOUNT*0 ; buffer1
			dc.b SPRITECOUNT*1 ; buffer2

spriteCodeSizeStart

;--------------------------------------------------------------------
; draws all the sprites (order is from front to back)
;--------------------------------------------------------------------
paintSprites SUBROUTINE
	ldx #$00
.yolo1
	;lda gameObjectUsed,x
	;beq .dontpaint1
	lda spriteVisible,x
	beq .dontpaint1
	lda spriteScreenSpace,x
	bne .dontpaint1
	stx .xrestore1 + 1
	jsr paintSprite
.xrestore1
	ldx #$44
.dontpaint1
	inx
	cpx #SPRITECOUNT
	bne .yolo1

	ldx #$00
.yolo2
	;lda gameObjectUsed,x
	;beq .dontpaint2
	lda spriteVisible,x
	beq .dontpaint2
	lda spriteScreenSpace,x
	beq .dontpaint2
	stx .xrestore2 + 1
	jsr paintSprite
.xrestore2
	ldx #$44
.dontpaint2
	inx
	cpx #SPRITECOUNT
	bne .yolo2

	rts

;--------------------------------------------------------------------
; clears all the sprites (order is from back to front)
;--------------------------------------------------------------------
clearSprites SUBROUTINE
	ldx doubleBuffer
	lda spriteRestoreIndexAdd,x
	sta .stopcmp + 1
	clc
	adc #SPRITECOUNT-1
	tax

.nextsprite
	lda spriteRestoreOn,x
	beq .norestore
	stx .xrestore + 1
	lda #$00
	sta spriteRestoreOn,x
	lda spriteRestoreXS,x
	sta rectLeft
	lda spriteRestoreYS,x
	sta rectTop
	lda spriteRestoreXE,x
	sta rectRight
	lda spriteRestoreYE,x
	sta rectBottom
	jsr restoreBackgroundRect
.done
.xrestore
	ldx #$44
.norestore
	dex
.stopcmp
	cpx #$44
	bpl .nextsprite
	rts

;--------------------------------------------------------------------
; resets all sprite restores (for instance for level switches)
;--------------------------------------------------------------------
restoreBackgroundRect SUBROUTINE
	rts

;--------------------------------------------------------------------
; resets all sprite restores (for instance for level switches)
;--------------------------------------------------------------------
clearSpriteRestores SUBROUTINE
	ldx #SPRITECOUNT*2-1
	lda #$00
.again
	sta spriteRestoreOn,x
	dex
	bpl .again
	rts

spriteScreenXSLo dc.b $00
spriteScreenXSHi dc.b $00
spriteScreenYSLo dc.b $00
spriteScreenYSHi dc.b $00
spriteScreenXSLo4 dc.b $00
spriteScreenYSLo8 dc.b $00
spriteScreenXELo4 dc.b $00
spriteScreenYELo8 dc.b $00
spriteScreenXELo dc.b $00
spriteScreenXEHi dc.b $00
spriteScreenYELo dc.b $00
spriteScreenYEHi dc.b $00
spriteScreenXSize dc.b $00
spriteScreenXSize4 dc.b $00
spriteScreenYSize dc.b $00
spriteScreenXChars dc.b $00
spriteScreenYChars dc.b $00
spriteScreenYCharsMul8 dc.b $00
spriteX3Ofs dc.b $00
spriteY7Ofs dc.b $00
spriteCurrentSpriteNr dc.b $00

spriteXPosLo2 dc.b $00
spriteXPosHi2 dc.b $00
spriteYPosLo2 dc.b $00
spriteYPosHi2 dc.b $00

;--------------------------------------------------------------------
; paints one sprite
;--------------------------------------------------------------------
paintSprite SUBROUTINE
	stx spriteCurrentSpriteNr

	lda spriteScreenSpace,x
	beq .normal
	lda scrollXLo
	and #%00000111
	sta .addxlo + 1
	lda scrollYLo
	and #%00000111
	sta .addylo + 1

	lda spriteXPosLo2
	clc
.addxlo
	adc #$44
	sta spriteScreenXSLo
	lda spriteXPosHi2
	adc #$00
	sta spriteScreenXSHi

	lda spriteYPosLo2
	clc
.addylo
	adc #$44
	sta spriteScreenYSLo
	lda spriteYPosHi2
	adc #$00
	sta spriteScreenYSHi
	jmp .donescrollvals
.normal

	lda spriteXPosLo,x
	sta spriteXPosLo2
	lda spriteXPosHi,x
	sta spriteXPosHi2
	lda spriteYPosLo,x
	sta spriteYPosLo2
	lda spriteYPosHi,x
	sta spriteYPosHi2

	; ---------------------------------------------

	; calculate direct screen position
	lda scrollXLo
	and #%11111000
	sta .subxlo + 1
	lda scrollYLo
	and #%11111000
	sta .subylo + 1

	lda spriteXPosLo2
	sec
.subxlo
	sbc #$44
	sta spriteScreenXSLo
	lda spriteXPosHi2
	sbc scrollXHi
	sta spriteScreenXSHi
	
	lda spriteYPosLo2
	sec
.subylo
	sbc #$44
	sta spriteScreenYSLo
	lda spriteYPosHi2
	sbc scrollYHi
	sta spriteScreenYSHi
.donescrollvals

	; x div 2
	lda spriteScreenXSHi
	lsr
	ror spriteScreenXSLo
	lda spriteScreenXSHi
	cmp #$80
	ror spriteScreenXSHi

	; do position raw clipping
	lda spriteScreenXSLo
	sec
	sbc #<[SCREENCOLS*4]
	lda spriteScreenXSHi
	sbc #>[SCREENCOLS*4]
	bmi .notright
	rts
.notright

	lda spriteScreenYSLo
	sec
	sbc #<[SCREENROWS*8]
	lda spriteScreenYSHi
	sbc #>[SCREENROWS*8]
	bmi .notdown
.rts2
	rts
.notdown

	lda spritePointer,x
	asl
	bcc .notabove127_1
	inc ZP_SPRITES + 1
.notabove127_1
	tay
	lda (ZP_SPRITES),y
	sta ZP_SPRITE + 0
	iny
	lda (ZP_SPRITES),y
	sta ZP_SPRITE + 1
	bcc .notabove127_2
	dec ZP_SPRITES + 1
.notabove127_2

	ldy #$00

	lda (ZP_SPRITE),y
	and #$03
	clc
	adc #$01
	sta spriteScreenXSize4 ; 1..4 chars
	asl
	asl
	sta spriteScreenXSize 
	lda (ZP_SPRITE),y
	lsr
	lsr
	sta spriteScreenYSize ; 0..63
	iny
	lda (ZP_SPRITE),y
	lsr ; 0 bit is internal sprite flipped flag
	tax
	and #%0000111 ; x add
	clc
	adc spriteScreenXSLo
	sta spriteScreenXSLo
	bcc .nohix
	inc spriteScreenXSHi
.nohix
	txa
	lsr
	lsr
	lsr
	clc
	adc spriteScreenYSLo
	sta spriteScreenYSLo
	bcc .nohiy
	inc spriteScreenYSHi
.nohiy
	iny
	tya ; setup sprite data for read
	clc
	adc ZP_SPRITE + 0
	sta ZP_SPRITEREAD + 0
	lda ZP_SPRITE + 1
	adc #$00
	sta ZP_SPRITEREAD + 1
	
	; ---
	ldx spriteCurrentSpriteNr

	; calculate screen ends and raw clip	
	lda spriteScreenXSLo
	clc
	adc spriteScreenXSize
	sta spriteScreenXELo
	lda spriteScreenXSHi
	adc #$00
	sta spriteScreenXEHi
	bpl .rawclip2x
	rts
.rawclip2x

	lda spriteScreenYSLo
	clc
	adc spriteScreenYSize
	sta spriteScreenYELo
	lda spriteScreenYSHi
	adc #$00
	sta spriteScreenYEHi
	bpl .rawclip2y
	rts
.rawclip2y

	; real screen positions
	; should work also for clipped
	lda spriteScreenXSLo
	and #%11111100
	sta spriteScreenXSLo4
	lda spriteScreenYSLo
	and #%11111000
	sta spriteScreenYSLo8
	lda spriteScreenXELo
	clc
	adc #$03
	and #%11111100
	sta spriteScreenXELo4
	lda spriteScreenYELo
	clc
	adc #$07
	and #%11111000
	sta spriteScreenYELo8

	; calculate sizes
	; should work also for clipped
	lda spriteScreenXELo4
	sec
	sbc spriteScreenXSLo4
	lsr
	lsr
	sta spriteScreenXChars
	
	lda spriteScreenYELo8
	sec
	sbc spriteScreenYSLo8
	sta spriteScreenYCharsMul8
	lsr
	lsr
	lsr
	sta spriteScreenYChars

	lda spriteScreenXSLo
	sec
	sbc spriteScreenXSLo4
	sta spriteX3Ofs

	; flipsprite if necessary
	ldy #$01
	ldx spriteCurrentSpriteNr
	lda spriteXFlipped,x
	eor (ZP_SPRITE),y
	and #$01
	beq .noflipneeded
	lda (ZP_SPRITE),y
	and #%11111110
	ora spriteXFlipped,x
	sta (ZP_SPRITE),y
	lda spriteScreenXSize4
	asl
	tax
	lda spriteFlipRoutines+0,x
	sta .flipJsr + 1
	lda spriteFlipRoutines+1,x
	sta .flipJsr + 2
.flipJsr
	jsr $4444
.noflipneeded

	lda spriteScreenYSLo
	sec
	sbc spriteScreenYSLo8
	sta spriteY7Ofs
;	beq .noyzerofillup ; you can implement this again if the sprites have heights of 8 or 16 and so on
	; clear surrounding sprite area
	lda spriteScreenXChars
	asl
	tax
	lda spriteClearRoutines+0,x
	sta .clearJsr + 1
	lda spriteClearRoutines+1,x
	sta .clearJsr + 2
.clearJsr
	jsr $4444
;.noyzerofillup

	lda ZP_SPRITECHARSETWRITE + 0
	clc
	adc spriteY7Ofs
	sta ZP_SPRITEWRITE + 0
	lda ZP_SPRITECHARSETWRITE + 1
	adc #$00
	sta ZP_SPRITEWRITE + 1

	lda spriteScreenXChars
	asl
	asl
	ora spriteX3Ofs
	asl
	tax
	lda spriteBlitRoutines + 0,x
	sta .blitJsr + 1
	lda spriteBlitRoutines + 1,x
	sta .blitJsr + 2
.blitJsr
	jsr $4444

	; calculate end char (for test if above 255 chars)
	lda #$00
	ldx spriteScreenYChars
	clc
.next
	adc spriteScreenXChars
	dex
	bne .next
	sta spriteEndCharCount
	adc ZP_SPRITECHARSETCHAR
	bcc .notabove255chars
	rts
.notabove255chars

	; check if all is in screen and not clipped
	lda spriteScreenXSHi
	ora spriteScreenYSHi
	bmi .clipped

	; do position raw clipping
	lda spriteScreenXELo
	sec
	sbc #<[SCREENCOLS*4]
	lda spriteScreenXEHi
	sbc #>[SCREENCOLS*4]
	bpl .clipped

	lda spriteScreenYELo
	sec
	sbc #<[SCREENROWS*8]
	lda spriteScreenYEHi
	sbc #>[SCREENROWS*8]
	bpl .clipped

	jmp paintSpriteNotClipped
.clipped
	jmp paintSpriteClipped

spriteEndCharCount dc.b $00

;--------------------------------------------------------------------
paintSpriteNotClipped
	lda spriteScreenYSLo8
	lsr
	lsr
	lsr
	sta .yp + 1
	tay
	lda spriteScreenXSLo4
	lsr
	lsr
	sta .xp + 1
	clc
	adc MUL40TABLO,y
	sta ZP_SPRITEWRITE + 0
	lda MUL40TABHI,y
	adc ZP_SCREENHI
	ora #>$400
	sta ZP_SPRITEWRITE + 1
	
	lda spriteScreenYChars
	clc
	adc ZP_SPRITECHARSETCHAR
	sta .yLoopCount + 1
	lda spriteScreenXChars
	sta .xLoopCount + 1
	lda #$00
	sta .xadr + 1
	
	ldx ZP_SPRITECHARSETCHAR
.nextx
.xadr
	ldy #$44
	clc
.nexty
	lda (ZP_SPRITEWRITE),y
	beq .emptyblock
	jsr spriteMergeBackground
.emptyblock
	txa
	sta (ZP_SPRITEWRITE),y
	tya
	adc #40
	tay
	inx
.yLoopCount
	cpx #$44
	bne .nexty 
	txa
	clc
	adc spriteScreenYChars
	sta .yLoopCount + 1
	inc .xadr + 1
	dec .xLoopCount + 1
.xLoopCount
	lda #$44
	bne .nextx

	; write restore structure
	ldx doubleBuffer
	lda spriteCurrentSpriteNr
	clc
	adc spriteRestoreIndexAdd,x
	tax
	lda #$01
	sta spriteRestoreOn,x
.xp
	lda #$44
	sta spriteRestoreXS,x
	clc
	adc spriteScreenXChars
	sta spriteRestoreXE,x
.yp
	lda #$44
	sta spriteRestoreYS,x
	clc
	adc spriteScreenYChars
	sta spriteRestoreYE,x

	jmp spriteDone

.yclipadder dc.b $00
.xclipadder dc.b $00
.yfulllineadd dc.b $00
;--------------------------------------------------------------------
paintSpriteClipped

	lda #$00
	sta .yclipadder
	lda spriteScreenYSLo8
	lsr
	lsr
	lsr
	sta .ypc + 1
	lda spriteScreenYSHi
	asl
	asl
	asl
	asl
	asl
	ora .ypc + 1 
	; we are on screen so no need for bigger coordinates
	bpl .notclippedup
	eor #$ff
	clc
	adc #$01
	sta .yclipadder
	lda #$00
.notclippedup
	sta .ypc + 1
	tay
	
	lda #$00
	sta .xclipadder
	lda spriteScreenXSLo4
	lsr
	lsr
	sta .xpc + 1
	lda spriteScreenXSHi
	asl
	asl
	asl
	asl
	asl
	asl
	ora .xpc + 1
	; we are on screen so no need for bigger coordinates
	bpl .notclippedleft
	eor #$ff
	clc
	adc #$01
	sta .xclipadder
	lda #$00
.notclippedleft
	sta .xpc + 1
	clc
	adc MUL40TABLO,y
	sta ZP_SPRITEWRITE + 0
	lda MUL40TABHI,y
	adc ZP_SCREENHI
	ora #>$400
	sta ZP_SPRITEWRITE + 1

	; screen start pos calculated
	
	lda spriteScreenYChars
	sec
	sbc .yclipadder
	clc
	adc .ypc + 1
	cmp #SCREENROWS
	bmi .notclippeddown
	lda #SCREENROWS
.notclippeddown
	sta .yec + 1
	sec
	sbc .ypc + 1
	sta .yLoopCountC + 1
	bmi .rts
	bne .doy
.rts
	rts
.doy

	lda spriteScreenYChars
	sec
	sbc .yLoopCountC + 1
	sta .yfulllineadd

	lda spriteScreenXChars
	sec
	sbc .xclipadder
	clc
	adc .xpc + 1
	cmp #SCREENCOLS
	bmi .notclippedright
	lda #SCREENCOLS
.notclippedright
	sta .xec + 1
	sec
	sbc .xpc + 1
	sta .xLoopCountC + 1
	bmi .rts
	beq .rts

	; write restore structure
	ldx doubleBuffer
	lda spriteCurrentSpriteNr
	clc
	adc spriteRestoreIndexAdd,x
	tax
	lda #$01
	sta spriteRestoreOn,x
.xpc
	lda #$44
	sta spriteRestoreXS,x
	clc
	adc spriteScreenXChars
.xec
	lda #$44
	sta spriteRestoreXE,x
.ypc
	lda #$44
	sta spriteRestoreYS,x
	clc
	adc spriteScreenYChars
.yec
	lda #$44
	sta spriteRestoreYE,x


	lda ZP_SPRITECHARSETCHAR
	ldx .xclipadder
	beq .nox
.zwo
	adc spriteScreenYChars
	dex
	bne .zwo
.nox
	clc
	adc .yclipadder
	tax
	clc
	adc .yLoopCountC + 1
	sta .yLoopCountC + 1

	lda #$00
	sta .xadrC + 1

.nextxC
.xadrC
	ldy #$44
	clc
.nextyC
	lda (ZP_SPRITEWRITE),y
	beq .emptyblockC
	jsr spriteMergeBackground
.emptyblockC
	txa
	sta (ZP_SPRITEWRITE),y
	tya
	adc #40
	tay
	inx
.yLoopCountC
	cpx #$44
	bne .nextyC 
	txa
	clc
	adc spriteScreenYChars
	sta .yLoopCountC + 1
	txa
	adc .yfulllineadd
	tax 
	inc .xadrC + 1
	dec .xLoopCountC + 1
.xLoopCountC
	lda #$44
	bne .nextxC

	jmp spriteDone

;--------------------------------------------------------------------
spriteMergeBackground SUBROUTINE
	stx .xrestore + 1
	sty .yrestore + 1
	tax
	lda MUL8TABLO,x
	sta ZP_SPRITECHARSETWRITE3 + 0 ; actually read
	lda MUL8TABHI,x
	clc
	adc ZP_CHARSETHI
	sta ZP_SPRITECHARSETWRITE3 + 1
	ldx .xrestore + 1
	lda MUL8TABLO,x
	sta ZP_SPRITECHARSETWRITE2 + 0 ; actually read
	lda MUL8TABHI,x
	adc ZP_CHARSETHI
	sta ZP_SPRITECHARSETWRITE2 + 1 
	ldy #$00
	REPEAT $07
	lax (ZP_SPRITECHARSETWRITE2),y ; sprite
	and (ZP_SPRITECHARSETWRITE3),y ; background and sprite
	ora automask,x
	sta (ZP_SPRITECHARSETWRITE2),y ; sprite
	iny
	REPEND
	lax (ZP_SPRITECHARSETWRITE2),y ; sprite
	and (ZP_SPRITECHARSETWRITE3),y ; background and sprite
	ora automask,x
	sta (ZP_SPRITECHARSETWRITE2),y ; sprite
.xrestore
	ldx #$44
.yrestore
	ldy #$44
	rts

spriteDone SUBROUTINE
	; calculate addition count
	lda spriteEndCharCount
	clc
	adc ZP_SPRITECHARSETCHAR
	sta ZP_SPRITECHARSETCHAR
	lda spriteEndCharCount
	asl
	asl
	asl
	clc
	adc ZP_SPRITECHARSETWRITE + 0
	sta ZP_SPRITECHARSETWRITE + 0
	bcc .nohi
	inc ZP_SPRITECHARSETWRITE + 1
.nohi ; the count of chars added should not exceed 32
	rts

spriteClearRoutines
	; (8 added because of positioning)
	dc.w rtsFunc
	dc.w spriteClearX1 ; 0-3 pixel
	dc.w spriteClearX2 ; 0-7 pixel
	dc.w spriteClearX3 ; 0-11 pixel 
	dc.w spriteClearX4 ; 0-15 pixel 
	dc.w spriteClearX5 ; 0-19 pixel 

spriteClearX1 SUBROUTINE
	ldy spriteY7Ofs
	dey
	bmi .no1
	lda #FILLUPBYTE
.nexty1
	sta (ZP_SPRITECHARSETWRITE),y
	dey
	bpl .nexty1
.no1
	lda spriteScreenYSize
	clc
	adc spriteY7Ofs
	tay
	sta .sub + 1
	lda spriteScreenYCharsMul8
	sec
.sub
	sbc #$44
	tax
	beq .no2
	bmi .no2
	lda #FILLUPBYTE
.nexty2
	sta (ZP_SPRITECHARSETWRITE),y
	iny
	dex
	bne .nexty2
.no2
	rts

spriteClearX2 SUBROUTINE
	lda ZP_SPRITECHARSETWRITE + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE2 + 0
	lda ZP_SPRITECHARSETWRITE + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE2 + 1

	ldy spriteY7Ofs
	dey
	bmi .no1
	lda #FILLUPBYTE
.nexty1
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	dey
	bpl .nexty1
.no1
	lda spriteScreenYSize
	clc
	adc spriteY7Ofs
	tay
	sta .sub + 1
	lda spriteScreenYCharsMul8
	sec
.sub
	sbc #$44
	tax
	beq .no2
	bmi .no2
	lda #FILLUPBYTE
.nexty2
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	iny
	dex
	bne .nexty2
.no2
	rts

spriteClearX3 SUBROUTINE
	lda ZP_SPRITECHARSETWRITE + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE2 + 0
	lda ZP_SPRITECHARSETWRITE + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE2 + 1
	lda ZP_SPRITECHARSETWRITE2 + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE3 + 0
	lda ZP_SPRITECHARSETWRITE2 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE3 + 1

	ldy spriteY7Ofs
	dey
	bmi .no1
	lda #FILLUPBYTE
.nexty1
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	sta (ZP_SPRITECHARSETWRITE3),y
	dey
	bpl .nexty1
.no1
	lda spriteScreenYSize
	clc
	adc spriteY7Ofs
	tay
	sta .sub + 1
	lda spriteScreenYCharsMul8
	sec
.sub
	sbc #$44
	tax
	beq .no2
	bmi .no2
	lda #FILLUPBYTE
.nexty2
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	sta (ZP_SPRITECHARSETWRITE3),y
	iny
	dex
	bne .nexty2
.no2
	rts

spriteClearX4 SUBROUTINE
	lda ZP_SPRITECHARSETWRITE + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE2 + 0
	lda ZP_SPRITECHARSETWRITE + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE2 + 1
	lda ZP_SPRITECHARSETWRITE2 + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE3 + 0
	lda ZP_SPRITECHARSETWRITE2 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE3 + 1
	lda ZP_SPRITECHARSETWRITE3 + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE4 + 0
	lda ZP_SPRITECHARSETWRITE3 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE4 + 1

	ldy spriteY7Ofs
	dey
	bmi .no1
	lda #FILLUPBYTE
.nexty1
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	sta (ZP_SPRITECHARSETWRITE3),y
	sta (ZP_SPRITECHARSETWRITE4),y
	dey
	bpl .nexty1
.no1
	lda spriteScreenYSize
	clc
	adc spriteY7Ofs
	tay
	sta .sub + 1
	lda spriteScreenYCharsMul8
	sec
.sub
	sbc #$44
	tax
	beq .no2
	bmi .no2
	lda #FILLUPBYTE
.nexty2
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	sta (ZP_SPRITECHARSETWRITE3),y
	sta (ZP_SPRITECHARSETWRITE4),y
	iny
	dex
	bne .nexty2
.no2
	rts
	
spriteClearX5 SUBROUTINE
	lda ZP_SPRITECHARSETWRITE + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE2 + 0
	lda ZP_SPRITECHARSETWRITE + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE2 + 1
	lda ZP_SPRITECHARSETWRITE2 + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE3 + 0
	lda ZP_SPRITECHARSETWRITE2 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE3 + 1
	lda ZP_SPRITECHARSETWRITE3 + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_SPRITECHARSETWRITE4 + 0
	lda ZP_SPRITECHARSETWRITE3 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE4 + 1
	lda ZP_SPRITECHARSETWRITE4 + 0
	clc
	adc spriteScreenYCharsMul8
	sta ZP_WRITE + 0
	lda ZP_SPRITECHARSETWRITE4 + 1
	adc #$00
	sta ZP_WRITE + 1

	ldy spriteY7Ofs
	dey
	bmi .no1
	lda #FILLUPBYTE
.nexty1
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	sta (ZP_SPRITECHARSETWRITE3),y
	sta (ZP_SPRITECHARSETWRITE4),y
	sta (ZP_WRITE),y
	dey
	bpl .nexty1
.no1
	lda spriteScreenYSize
	clc
	adc spriteY7Ofs
	tay
	sta .sub + 1
	lda spriteScreenYCharsMul8
	sec
.sub
	sbc #$44
	tax
	beq .no2
	bmi .no2
	lda #FILLUPBYTE
.nexty2
	sta (ZP_SPRITECHARSETWRITE),y
	sta (ZP_SPRITECHARSETWRITE2),y
	sta (ZP_SPRITECHARSETWRITE3),y
	sta (ZP_SPRITECHARSETWRITE4),y
	sta (ZP_WRITE),y
	iny
	dex
	bne .nexty2
.no2
	rts

spriteBlitRoutines
	dc.w rtsFunc
	dc.w rtsFunc
	dc.w rtsFunc
	dc.w rtsFunc
	dc.w blitSpriteDirect1
	dc.w blitSpriteShifted1
	dc.w blitSpriteShifted1
	dc.w blitSpriteShifted1
	dc.w blitSpriteDirect2
	dc.w blitSpriteShifted2
	dc.w blitSpriteShifted2
	dc.w blitSpriteShifted2
	dc.w blitSpriteDirect3
	dc.w blitSpriteShifted3
	dc.w blitSpriteShifted3
	dc.w blitSpriteShifted3
	dc.w blitSpriteDirect4
	dc.w blitSpriteShifted4
	dc.w blitSpriteShifted4
	dc.w blitSpriteShifted4
	dc.w blitSpriteDirect5
	dc.w blitSpriteShifted5
	dc.w blitSpriteShifted5
	dc.w blitSpriteShifted5

	MAC spriteAdvance
	lda {1} + 0
	adc spriteScreenYSize
	sta {2} + 0
	lda {1} + 1
	adc #$00
	sta {2} + 1
	lda {3} + 0
	adc spriteScreenYCharsMul8
	sta {4} + 0
	lda {3} + 1
	adc #$00
	sta {4} + 1
	ENDM

blitSpriteDirect1 SUBROUTINE
	ldy spriteScreenYSize
	dey
.next
	lda (ZP_SPRITEREAD),y
	sta (ZP_SPRITEWRITE),y
	dey
	bpl .next
	rts

blitSpriteDirect2 SUBROUTINE
	clc
	spriteAdvance ZP_SPRITEREAD,[dmodread1a+1],ZP_SPRITEWRITE,[dmodwrite1a+1]
	ldy spriteScreenYSize
	dey
.next
	lda (ZP_SPRITEREAD),y
	sta (ZP_SPRITEWRITE),y
dmodread1a
	lda $4444,y
dmodwrite1a
	sta $4444,y
	dey
	bpl .next
	rts

blitSpriteDirect3 SUBROUTINE
	clc
	spriteAdvance ZP_SPRITEREAD,[dmodread1b+1],ZP_SPRITEWRITE,[dmodwrite1b+1]
	spriteAdvance [dmodread1b+1],[dmodread2b+1],[dmodwrite1b+1],[dmodwrite2b+1]
	ldy spriteScreenYSize
	dey
.next
	lda (ZP_SPRITEREAD),y
	sta (ZP_SPRITEWRITE),y
dmodread1b
	lda $4444,y
dmodwrite1b
	sta $4444,y
dmodread2b
	lda $4444,y
dmodwrite2b
	sta $4444,y
	dey
	bpl .next
	rts

blitSpriteDirect4 SUBROUTINE
	clc
	spriteAdvance ZP_SPRITEREAD,[emodread1b+1],ZP_SPRITEWRITE,[emodwrite1b+1]
	spriteAdvance [emodread1b+1],[emodread2b+1],[emodwrite1b+1],[emodwrite2b+1]
	spriteAdvance [emodread2b+1],[emodread3b+1],[emodwrite2b+1],[emodwrite3b+1]
	ldy spriteScreenYSize
	dey
.next
	lda (ZP_SPRITEREAD),y
	sta (ZP_SPRITEWRITE),y
emodread1b
	lda $4444,y
emodwrite1b
	sta $4444,y
emodread2b
	lda $4444,y
emodwrite2b
	sta $4444,y
emodread3b
	lda $4444,y
emodwrite3b
	sta $4444,y
	dey
	bpl .next
	rts

blitSpriteDirect5 SUBROUTINE
	inc $ff15
	rts

blitSpriteShifted1 SUBROUTINE
.error
	inc $ff19 ; one character is never shifted (it's two then)
	jmp .error

blitSpriteShifted2 SUBROUTINE
	clc
	spriteAdvance [smodtemp],[smodtemp],ZP_SPRITEWRITE,[smodwrite1a+1]
	lda spriteX3Ofs
	adc #[>rorTablesFillup]-1
	sta .tab1 + 2
	lda spriteX3Ofs
	adc #[>aslTablesFillup]-1
	sta .tab2a + 2
	ldy spriteScreenYSize
	dey
.next
	lax (ZP_SPRITEREAD),y
.tab1
	lda $4400,x
	sta (ZP_SPRITEWRITE),y
.tab2a
	lda $4400,x
smodwrite1a
	sta $4444,y
	dey
	bpl .next
	rts

blitSpriteShifted3 SUBROUTINE
	clc
	spriteAdvance ZP_SPRITEREAD,[smodread1b+1],ZP_SPRITEWRITE,[smodwrite1b+1]
	spriteAdvance [smodtemp],[smodtemp],[smodwrite1b+1],[smodwrite2b+1]
	lda spriteX3Ofs
	tax
	adc #[>rorTablesFillup]-1
	sta .tab1 + 2
	txa
	adc #[>rorTables]-1
	sta .tab2b + 2
	txa
	adc #[>aslTables]-1
	sta .tab2a + 2
	txa
	adc #[>aslTablesFillup]-1
	sta .tab3a + 2
	ldy spriteScreenYSize
	dey
.next
	lax (ZP_SPRITEREAD),y
.tab1
	lda $4400,x
	sta (ZP_SPRITEWRITE),y
.tab2a
	lda $4400,x
smodread1b
	ldx $4444,y
.tab2b
	ora $4400,x
smodwrite1b
	sta $4444,y
.tab3a
	lda $4400,x
smodwrite2b
	sta $4444,y
	dey
	bpl .next
	rts

smodtemp dc.w $0000

blitSpriteShifted4 SUBROUTINE
	clc
	spriteAdvance ZP_SPRITEREAD,[smodread1c+1],ZP_SPRITEWRITE,[smodwrite1c+1]
	spriteAdvance [smodread1c+1],[smodread2c+1],[smodwrite1c+1],[smodwrite2c+1]
	spriteAdvance [smodtemp],[smodtemp],[smodwrite2c+1],[smodwrite3c+1]
	lda spriteX3Ofs
	tax
	adc #[>rorTablesFillup]-1
	sta .tab1 + 2
	txa
	adc #[>rorTables]-1
	sta .tab2b + 2
	sta .tab3b + 2
	txa
	adc #[>aslTables]-1
	sta .tab2a + 2
	sta .tab3a + 2
	txa
	adc #[>aslTablesFillup]-1
	sta .tab4a + 2
	ldy spriteScreenYSize
	dey
.next
	lax (ZP_SPRITEREAD),y
.tab1
	lda $4400,x
	sta (ZP_SPRITEWRITE),y
.tab2a
	lda $4400,x
smodread1c
	ldx $4444,y
.tab2b
	ora $4400,x
smodwrite1c
	sta $4444,y
.tab3a
	lda $4400,x
smodread2c
	ldx $4444,y
.tab3b
	ora $4400,x
smodwrite2c
	sta $4444,y
.tab4a
	lda $4400,x
smodwrite3c
	sta $4444,y
	dey
	bpl .next
	rts

blitSpriteShifted5 SUBROUTINE
	clc
	spriteAdvance ZP_SPRITEREAD,[smodread1d+1],ZP_SPRITEWRITE,[smodwrite1d+1]
	spriteAdvance [smodread1d+1],[smodread2d+1],[smodwrite1d+1],[smodwrite2d+1]
	spriteAdvance [smodread2d+1],[smodread3d+1],[smodwrite2d+1],[smodwrite3d+1]
	spriteAdvance [smodtemp],[smodtemp],[smodwrite3d+1],[smodwrite4d+1]
	lda spriteX3Ofs
	tax
	adc #[>rorTablesFillup]-1
	sta .tab1 + 2
	txa
	adc #[>rorTables]-1
	sta .tab2b + 2
	sta .tab3b + 2
	sta .tab4b + 2
	txa
	adc #[>aslTables]-1
	sta .tab2a + 2
	sta .tab3a + 2
	sta .tab4a + 2
	txa
	adc #[>aslTablesFillup]-1
	sta .tab5a + 2
	ldy spriteScreenYSize
	dey
.next

	lax (ZP_SPRITEREAD),y
.tab1
	lda $4400,x
	sta (ZP_SPRITEWRITE),y

.tab2a
	lda $4400,x
smodread1d
	ldx $4444,y
.tab2b
	ora $4400,x
smodwrite1d
	sta $4444,y

.tab3a
	lda $4400,x
smodread2d
	ldx $4444,y
.tab3b
	ora $4400,x
smodwrite2d
	sta $4444,y

.tab4a
	lda $4400,x
smodread3d
	ldx $4444,y
.tab4b
	ora $4400,x
smodwrite3d
	sta $4444,y

.tab5a
	lda $4400,x
smodwrite4d
	sta $4444,y
	dey
	bpl .next
	rts

spriteFlipRoutines
	dc.w rtsFunc
	dc.w spriteXFlip1
	dc.w spriteXFlip2
	dc.w spriteXFlip3
	dc.w spriteXFlip4

spriteXFlip1 SUBROUTINE
	ldy spriteScreenYSize
	dey
.nexty
	lax (ZP_SPRITEREAD),y
	lda xfliptable,x
	sta (ZP_SPRITEREAD),y
	dey
	bpl .nexty
	rts

spriteXFlip2 SUBROUTINE
	ldy spriteScreenYSize
	tya
	clc
	adc ZP_SPRITEREAD + 0
	sta ZP_SPRITECHARSETWRITE2 + 0
	lda ZP_SPRITEREAD + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE2 + 1
	dey
.nexty
	lax (ZP_SPRITEREAD),y
	lda xfliptable,x
	sta .write2 + 1
	lax (ZP_SPRITECHARSETWRITE2),y
	lda xfliptable,x
	sta (ZP_SPRITEREAD),y
.write2
	lda #$44
	sta (ZP_SPRITECHARSETWRITE2),y
	dey
	bpl .nexty
	rts

spriteXFlip3 SUBROUTINE
	ldy spriteScreenYSize
	tya
	clc
	adc ZP_SPRITEREAD + 0
	sta ZP_SPRITECHARSETWRITE2 + 0
	lda ZP_SPRITEREAD + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE2 + 1
	tya
	clc
	adc ZP_SPRITECHARSETWRITE2 + 0
	sta ZP_SPRITECHARSETWRITE3 + 0
	lda ZP_SPRITECHARSETWRITE2 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE3 + 1
	dey
.nexty
	lax (ZP_SPRITECHARSETWRITE2),y
	lda xfliptable,x
	sta (ZP_SPRITECHARSETWRITE2),y
	lax (ZP_SPRITEREAD),y
	lda xfliptable,x
	sta .write2 + 1
	lax (ZP_SPRITECHARSETWRITE3),y
	lda xfliptable,x
	sta (ZP_SPRITEREAD),y
.write2
	lda #$44
	sta (ZP_SPRITECHARSETWRITE3),y
	dey
	bpl .nexty
	rts

spriteXFlip4 SUBROUTINE
	ldy spriteScreenYSize
	tya
	clc
	adc ZP_SPRITEREAD + 0
	sta ZP_SPRITECHARSETWRITE2 + 0
	lda ZP_SPRITEREAD + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE2 + 1
	tya
	clc
	adc ZP_SPRITECHARSETWRITE2 + 0
	sta ZP_SPRITECHARSETWRITE3 + 0
	lda ZP_SPRITECHARSETWRITE2 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE3 + 1
	tya
	clc
	adc ZP_SPRITECHARSETWRITE3 + 0
	sta ZP_SPRITECHARSETWRITE4 + 0
	lda ZP_SPRITECHARSETWRITE3 + 1
	adc #$00
	sta ZP_SPRITECHARSETWRITE4 + 1
	dey
.nexty
	lax (ZP_SPRITEREAD),y
	lda xfliptable,x
	sta .write1 + 1
	lax (ZP_SPRITECHARSETWRITE4),y
	lda xfliptable,x
	sta (ZP_SPRITEREAD),y
.write1
	lda #$44
	sta (ZP_SPRITECHARSETWRITE4),y

	lax (ZP_SPRITECHARSETWRITE2),y
	lda xfliptable,x
	sta .write2 + 1
	lax (ZP_SPRITECHARSETWRITE3),y
	lda xfliptable,x
	sta (ZP_SPRITECHARSETWRITE2),y
.write2
	lda #$44
	sta (ZP_SPRITECHARSETWRITE3),y
	dey
	bpl .nexty
	rts

;--------------------------------------------------------------------
; initializing the sprite subsystem
;--------------------------------------------------------------------
initSprites SUBROUTINE
	lda #<spriteStart
	sta ZP_SPRITES + 0
	lda #>spriteStart
	sta ZP_SPRITES + 1
	rts

rtsFunc SUBROUTINE
	rts

spriteCodeSizeEnd

