//--------------------------------------------------------------------------//
// iq / rgba  .  tiny codes  .  2008                                        //
//--------------------------------------------------------------------------//
#define _CRT_SECURE_NO_WARNINGS
#pragma comment( lib, "winmm.lib" )
#include "windows.h"
#include <math.h>
#include <stdlib.h>
#include <stdio.h>
#include <conio.h>
#include <string>
#include <vector>
#include <set>
#include <algorithm>
#include <map>
#include "png/png.h"
#include "setup.h"
#include <fstream>

unsigned int palette[256];

unsigned char getColor(unsigned int color) 
{
	int r = color & 255;
	int g = (color>>8) & 255;
	int b = (color>>16) & 255;
	int lastDelta = 10000*10000;
	int ri = 0;
	unsigned int colorEnabled[16] = {1,1,1,1,1,1,1,1, 1,1,1,1,1,1,1,1};

	for (int i = 0; i < 128; ++i) 
	{
		int cr = (palette[i]>>16) & 255;
		int cg = (palette[i]>>8) & 255;
		int cb = palette[i] & 255;
		int dr = cr - r;
		int dg = cg - g;
		int db = cb - b;
		int d = dr * dr + dg * dg + db * db;

		if (d < lastDelta && (colorEnabled[i & 15] != 0)) 
		{
			lastDelta = d;
			ri = i;
		}
	}
	if ((ri & 0x0f)==0x00)
		ri = 0;
	return ri;
}

unsigned char getColor8(int color) 
{
	int r = color & 255;
	int g = (color>>8) & 255;
	int b = (color>>16) & 255;
	int lastDelta = 10000*10000;
	int ri = 0;
	unsigned int colorEnabled[16] = {1,1,1,1,1,1,1,1, 0,0,0,0,0,0,0,0};

	for (int i = 0; i < 128; ++i) 
	{
		int cr = (palette[i]>>16) & 255;
		int cg = (palette[i]>>8) & 255;
		int cb = palette[i] & 255;
		int dr = cr - r;
		int dg = cg - g;
		int db = cb - b;
		int d = dr * dr + dg * dg + db * db;

		if (d < lastDelta && (colorEnabled[i & 15] != 0)) 
		{
			lastDelta = d;
			ri = i;
		}
	}
	if ((ri & 0x0f)==0x00)
		ri = 0;
	return ri;
}

#define MAXWIDTH 4096
#define MAXHEIGHT 4096

unsigned char bmpBuffer2[MAXWIDTH * MAXHEIGHT * 4 + 0x36];
int picture[MAXWIDTH*MAXHEIGHT];
int pictureS[MAXWIDTH*MAXHEIGHT];
int pictureWidth;
int pictureHeight;

extern "C"
{
#if defined(HAVE_LIBPNG) && defined(HAVE_LIBZ)
#  include <zlib.h>
#  ifdef HAVE_PNG_H
#    include <png.h>
#  else
#    include <libpng/png.h>
#  endif // HAVE_PNG_H
#endif // HAVE_LIBPNG && HAVE_LIBZ
}


/**
  The constructor loads the named PNG image from the given png filename.
  <P>The destructor free all memory and server resources that are used by
  the image.
*/
void LoadPNG(const char *png) // I - File to read
{
  int		i;			// Looping var
  FILE		*fp;			// File pointer
  int		channels;		// Number of color channels
  png_structp	pp;			// PNG read pointer
  png_infop	info;			// PNG info pointers
  png_bytep	*rows;			// PNG row pointers


  // Open the PNG file...
  if ((fp = fopen(png, "rb")) == NULL) {
	  int debug = 1;
	  Log("- load PNG failed:[%s]", png);
	  return;
  }

  // Setup the PNG data structures...
  pp   = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  info = png_create_info_struct(pp);

  if (setjmp(pp->jmpbuf))
  {
	  int debug = 1;
	  return;
  }

  // Initialize the PNG read "engine"...
  png_init_io(pp, fp);

  // Get the image dimensions and convert to grayscale or RGB...
  png_read_info(pp, info);

  if (info->color_type == PNG_COLOR_TYPE_PALETTE)
    png_set_expand(pp);

  if (info->color_type & PNG_COLOR_MASK_COLOR)
    channels = 3;
  else
    channels = 1;

  if ((info->color_type & PNG_COLOR_MASK_ALPHA) || info->num_trans)
    channels ++;

  int w = (int)(info->width);
  int h = (int)(info->height);
  int d = channels;
  pictureWidth = w;
  pictureHeight = h;

  if (info->bit_depth < 8)
  {
    png_set_packing(pp);
    png_set_expand(pp);
  }
  else if (info->bit_depth == 16)
    png_set_strip_16(pp);

#  if defined(HAVE_PNG_GET_VALID) && defined(HAVE_PNG_SET_TRNS_TO_ALPHA)
  // Handle transparency...
  if (png_get_valid(pp, info, PNG_INFO_tRNS))
    png_set_tRNS_to_alpha(pp);
#  endif // HAVE_PNG_GET_VALID && HAVE_PNG_SET_TRNS_TO_ALPHA

  unsigned char *array = (unsigned char *)pictureS;

  // Allocate pointers...
  rows = new png_bytep[h];

  for (i = 0; i < h; i ++)
    rows[i] = (png_bytep)(array + i * w * d); // we flip it

  // Read the image, handling interlacing as needed...
  for (i = png_set_interlace_handling(pp); i > 0; i --)
    png_read_rows(pp, rows, NULL, h);

#ifdef WIN32
  // Some Windows graphics drivers don't honor transparency when RGB == white
  if (channels == 4) 
  {
    // Convert RGB to 0 when alpha == 0...
    unsigned char *ptr = (unsigned char *)array;
    for (i = w * h; i > 0; i --, ptr += 4)
      if (!ptr[3]) ptr[0] = ptr[1] = ptr[2] = 0;
  }
#endif // WIN32

  if (channels == 3)
  {
	  unsigned char *array2 = new unsigned char[pictureWidth * pictureHeight * 4];
	  for (int i = w * h - 1; i >= 0; --i)
	  {
		  array2[i*4+0] = array[i*3+0];
		  array2[i*4+1] = array[i*3+1];
		  array2[i*4+2] = array[i*3+2];
		  array2[i*4+3] = 255;
	  }
	  memcpy(array, array2, w * h * 4);
	  delete[] array2;
  }

  // Free memory and return...
  delete[] rows;

  png_read_end(pp, info);
  png_destroy_read_struct(&pp, &info, NULL);

  fclose(fp);
}

void generateTables()
{
    std::string path = globalPath + "/data/output/";

	FILE *out;
//	for (int i = 0; i < 3; ++i)
//	{
//		char buffer[1000];
//		sprintf(buffer,(path+"asltable_%i.bin").c_str(),(i+1)*2);
//		out = fopen(buffer,"wb");
//		for (int j = 0; j < 256; ++j)
//		{
//			unsigned char c = j << ((i+1) * 2);
//			fwrite(&c,1,1,out);
//		}
//		fclose(out);
//
//		sprintf(buffer,(path+"rortable_%i.bin").c_str(),(3-i)*2);
//		out = fopen(buffer,"wb");
//		for (int j = 0; j < 256; ++j)
//		{
//			unsigned char c = j >> ((3-(i)) * 2);
//			fwrite(&c,1,1,out);
//		}
//		fclose(out);
//
//   		sprintf(buffer,(path+"asltable_%i_f.bin").c_str(),(i+1)*2);
//		out = fopen(buffer,"wb");
//		for (int j = 0; j < 256; ++j)
//		{
//			unsigned char c = j << ((i+1) * 2);
//            c |= 0xff >> (8-((i+1) * 2));
//            fwrite(&c,1,1,out);
//		}
//		fclose(out);
//
//		sprintf(buffer,(path+"rortable_%i_f.bin").c_str(),(3-i)*2);
//		out = fopen(buffer,"wb");
//		for (int j = 0; j < 256; ++j)
//		{
//			unsigned char c = j >> ((3-(i)) * 2);
//            c |= 0xff << (8-(3-(i)) * 2);
//			fwrite(&c,1,1,out);
//		}
//		fclose(out);
//    }
//
//	out = fopen((path+"automask.bin").c_str(),"wb");
//	for(int i = 0; i < 256; ++i)
//	{
//		unsigned char mask = 0;
//		for (int j = 0; j < 4; ++j)
//		{
//			int k = (i >> (j * 2)) & 3;
//			if (k != 3)
//				mask |= k << (j * 2);
//            else
//				mask |= 0 << (j * 2);
//		}
//		fwrite(&mask,1,1,out);
//	}
//	fclose(out);
//
// 	out = fopen((path+"xfliptable.bin").c_str(),"wb");
//	for(int i = 0; i < 256; ++i)
//	{
//		unsigned char c = 0;
//		for (int j = 0; j < 4; ++j)
//		{
//			int k = (i >> (j * 2)) & 3;
//		    c |= k << (6 - j * 2);
//        }
//		fwrite(&c,1,1,out);
//	}
//	fclose(out);
 	out = fopen((path+"sintab.bin").c_str(),"wb");
	for (int i = 0; i < 256; ++i) {
		int ci = sin((float)i / 256.f * 2.f * 3.14159f) * 128 + 128;
		if (ci < 0) ci = 0;
		if (ci > 255) ci = 255;
		unsigned char c = ci;
		fwrite(&c,1,1,out);
	}
	fclose(out);
}

struct SPRITE {
    SPRITE(const std::string &nam, const std::string &img, int x,int y, int xs, int ys, int xadd=0, int yadd=0, int blackCol = 0x00, int middleCol = 0x41, int whiteCol = 0x71)
    {
		name=nam;
		image = img;
		xpos=x;
        ypos=y;
        xsize=xs;
        ysize=ys;
		xa=xadd;
        ya=yadd;
		blackColor = blackCol;
		middleColor = middleCol;
		whiteColor = whiteCol;
	}
    std::string name;
    std::string image;
    int xpos,ypos,xsize,ysize;
    int xa,ya;
	int blackColor;
	int middleColor;
	int whiteColor;
};

void convertSprites(std::vector<SPRITE> sprites, const std::string &prefix) {
    std::string path = globalPath + "/data/output/";
    FILE *out = fopen((path + prefix + ".def").c_str(),"wb");
    int spriteCount = sprites.size();
    for (int i = 0; i < spriteCount; ++i) {
        fprintf(out,"SPR_%s = %d\n",sprites[i].name.c_str(), i);
    }
	fclose(out);
	out = fopen((path + prefix + ".inc").c_str(), "wb");
    fprintf(out,"spritePointers\n");
    for (int i = 0; i < spriteCount; ++i) {
		fprintf(out, "\tdc.w spriteData%d ; SPR_%s\n", i, sprites[i].name.c_str());
    }
    fprintf(out,"\n");
	for (int i = 0; i < spriteCount; ++i) {
		LoadPNG(sprites[i].image.c_str());
		SPRITE &sp = sprites[i];
		int bc = sp.blackColor;
		int mc = sp.middleColor;
		int white = sp.whiteColor;
		int mulX = 2;
		int mulY = 1;
		int spriteXSize = (sp.xsize + 7) / 8;
		int spriteYSize = sp.ysize;
		int spriteFlipped = 0;
		int xofs = sp.xpos;
		int yofs = sp.ypos;
		if (yofs + spriteYSize > pictureHeight)
			spriteYSize = pictureHeight-yofs;
		if (spriteYSize < 0)
			continue;
		int yposadd = sp.ya;
		if (yposadd > 15)
		{
			yposadd = 15;
		}
		fprintf(out, "\n\n\nspriteData%d ; SPR_%s\n", i, sp.name.c_str());
		fprintf(out, "\tdc.b %d + %d*4; xsize %d ysize %d\n", (spriteXSize - 1) , spriteYSize, spriteXSize, spriteYSize);
		fprintf(out, "\tdc.b %d + %d*2 + %d*16; flipped %d xadd %d yadd %d\n", (spriteFlipped ? 1 : 0) , sp.xa, yposadd, spriteFlipped, sp.xa, yposadd);

		std::set<int> colorsUsed;
		int q = 0;
		for (int x2 = 0; x2 < spriteXSize; ++x2) {
            for (int y = 0; y < spriteYSize; ++y) {
                unsigned char c = 0;
                for (int x = 0; x < 4; ++x) {
					int p = pictureS[((x+x2*4)+xofs/2)*mulX + (y+yofs)*mulY*pictureWidth];
					int q = getColor(p);
					colorsUsed.insert(q);
					int k;
					if (q == white)
						k = 2;
					else
					if (q == bc)
						k = 0;
					else
					if (q == mc)
						k = 1;
					else
						k = 3;
                    c |= k << (6-x*2);
                }
                if ((q & 15) == 0)
                    fprintf(out,"\n\tdc.b ");
                else 
                    fprintf(out,",");
                fprintf(out,"%d",c);
                q++;
            }
        }
    }
    fclose(out);
}

std::vector<std::string> split(const std::string& s, char seperator)
{
	std::vector<std::string> output;
	std::string::size_type prev_pos = 0, pos = 0;
	while ((pos = s.find(seperator, pos)) != std::string::npos)
	{
		std::string substring(s.substr(prev_pos, pos - prev_pos));
		output.push_back(substring);
		prev_pos = ++pos;
	}
	output.push_back(s.substr(prev_pos, pos - prev_pos)); // Last word
	return output;
}

void main( void )
{
	setupWithCommandLine();

	std::ifstream file(globalPath+"/data/palette/Plus4ColsMeasured.asc");
	std::string zeile;
	int nr = 0;
	while (std::getline(file, zeile) && zeile != "") {
		std::vector<std::string> values = split(zeile, ' ');
		if (values.size() == 3) {
			int r = atoi(values[0].c_str());
			int g = atoi(values[1].c_str());
			int b = atoi(values[2].c_str());
			palette[nr] = b | (g << 8) | (r << 16) | 0xff000000;
			nr++;
		}
	}
	file.close();

	generateTables();

	std::vector<SPRITE> sprites;
	std::string path = globalPath + "/data/sprites/";
	std::string luca1 = path + "luca1.png";

	sprites.push_back(SPRITE("MARIOWALK0", luca1, (5 + 3 * 0) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOWALK1", luca1, (5 + 3 * 1) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOWALK2", luca1, (5 + 3 * 2) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOJUMPUP", luca1, (5 + 3 * 4-1-1) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOJUMPFALL", luca1, (5 + 3 * 5 - 1 - 1) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIODEATH0", luca1, (5 + 3 * 6 - 1) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOCLIMB0", luca1, (5 + 3 * 6 + 4 * 2 - 1) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOCLIMB1", luca1, (5 + 3 * 7 + 4 * 2 - 1) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOSHOOTING", luca1, (5 + 3 * 6 + 4 * 1 - 1) * 8, (1 + 4 * 0) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIONOTE0", luca1, (5 + 3 * 6 + 1 + 2 * 0 - 1) * 8, (1 + 4 * 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIONOTE1", luca1, (5 + 3 * 6 + 1 + 2 * 1 - 1) * 8, (1 + 4 * 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIONOTE2", luca1, (5 + 3 * 6 + 1 + 2 * 2 - 1) * 8, (1 + 4 * 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIONOTE3", luca1, (5 + 3 * 6 + 1 + 2 * 3 - 1) * 8, (1 + 4 * 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIONOTE4", luca1, (5 + 3 * 6 + 1 + 2 * 4 - 1) * 8, (1 + 4 * 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIONOTE5", luca1, (5 + 3 * 6 + 1 + 2 * 5 - 1) * 8, (1 + 4 * 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOSWIM0", luca1, (5 + 3 * 4 - 1 - 1) * 8, (1 + 4 * 1) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOSWIM1", luca1, (5 + 3 * 5 - 1 - 1) * 8, (1 + 4 * 1) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOSWIM2", luca1, (5 + 3 * 6 - 1 - 1) * 8, (1 + 4 * 1) * 8, 16, 24, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOCROUCH0", luca1, (5 + 3 * 0) * 8, (1 + 4 * 1) * 8 + 8, 16, 16, 0, 8, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOCROUCH1", luca1, (5 + 3 * 1) * 8, (1 + 4 * 1) * 8 + 8, 16, 16, 0, 8, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("MARIOCROUCH2", luca1, (5 + 3 * 2) * 8, (1 + 4 * 1) * 8 + 8, 16, 16, 0, 8, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("KCAT0", luca1, (5 + 3 * 6 + 4 * 2 - 1 + 2 * 0) * 8, (6) * 8, 16, 16, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("KCAT1", luca1, (5 + 3 * 6 + 4 * 2 - 1 + 2 * 1) * 8, (6) * 8, 16, 16, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("KCAT2", luca1, (5 + 3 * 6 + 4 * 2 - 1 + 2 * 2) * 8, (6) * 8, 16, 16, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("KFISH0", luca1, (5 + 3 * 6 + 1 + 2 * 0-1) * 8, (1 + 4 * 1 + 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("KFISH1", luca1, (5 + 3 * 6 + 1 + 2 * 1-1) * 8, (1 + 4 * 1 + 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));
	sprites.push_back(SPRITE("KFISH2", luca1, (5 + 3 * 6 + 1 + 2 * 2-1) * 8, (1 + 4 * 1 + 1) * 8, 16, 8, 0, 0, 0x00, 0x5b, 0x71));

	convertSprites(sprites, "sprites1");
}

